//## begin module%1.4%.codegen_version preserve=yes
//   Read the documentation to learn more about C++ code generator
//   versioning.
//## end module%1.4%.codegen_version

//## begin module%41E6D10701BE.cm preserve=no
//	  %X% %Q% %Z% %W%
//## end module%41E6D10701BE.cm

//## begin module%41E6D10701BE.cp preserve=no
//## end module%41E6D10701BE.cp

//## Module: Histogram%41E6D10701BE; Package specification
//## Subsystem: XSUtil::Numerics%41E6D0F3003C
//## Source file: /local/data/xerxes1/gordon/headas/Xspec/src/XSUtil/Numerics/Histogram.h

#ifndef HISTOGRAM_H
#define HISTOGRAM_H 1

//## begin module%41E6D10701BE.additionalIncludes preserve=no
//## end module%41E6D10701BE.additionalIncludes

//## begin module%41E6D10701BE.includes preserve=yes
#include <xsTypes.h>
#include <XSUtil/Error/Error.h>
#include <cmath>
//## end module%41E6D10701BE.includes

//## begin module%41E6D10701BE.declarations preserve=no
//## end module%41E6D10701BE.declarations

//## begin module%41E6D10701BE.additionalDeclarations preserve=yes
//## end module%41E6D10701BE.additionalDeclarations


//## Modelname: XSUtil::Numerics%3F6F5A7603D4
namespace Numerics {
    //## begin Numerics%3F6F5A7603D4.initialDeclarations preserve=yes
    //## end Numerics%3F6F5A7603D4.initialDeclarations

    //## begin Numerics::Histogram%41E6D2050159.preface preserve=yes
    class LinearBinning
    {
       public:
          template <typename T>
          static int calcIndex(const T& min, const T& max, const T& val, const int nSteps,
                                const bool parIsLog = false)
          {
             // parIsLog flag is ignored for this case
             Real delta = (max - min)/static_cast<Real>(nSteps);
             return static_cast<int>(floor((val - min)/delta));
          }
          template <typename T>
          static void calcBoundaries(const T& min, const T& max, const int idx, 
                const int nSteps, std::pair<Real,Real>& boundaries, const bool parIsLog = false)
          {
             Real delta = (max - min)/static_cast<Real>(nSteps);
             boundaries.first = min + idx*delta;
             boundaries.second = min + (idx+1.0)*delta;
          }
          template <typename T>
          static int verify(const std::vector<T>& mins, const std::vector<T>& maxes, 
                        const BoolArray& isLog = BoolArray())
          {
             for (size_t i=0; i<mins.size(); ++i)
             {
                if (maxes[i] == mins[i]) return -2;
             }
             return 0;
          }
    };

    class LogBinning
    {
       public:
          template <typename T>
          static int calcIndex(const T& min, const T& max, const T& val, const int nSteps,
                                const bool parIsLog = true)
          {
             // parIsLogflag is ignored for this case
             using namespace std;
             const int BAD = -9999;
             if (val <= .0)  return BAD;
             Real delta = (log(max) - log(min))/static_cast<Real>(nSteps);
             return static_cast<int>(floor((log(val) - log(min))/delta));
          }
          template <typename T>
          static void calcBoundaries(const T& min, const T& max, const int idx, 
                const int nSteps, std::pair<Real,Real>& boundaries, const bool parIsLog = true)
          {
             using namespace std;
             Real delta = (log(max) - log(min))/static_cast<Real>(nSteps);
             boundaries.first = min*exp(idx*delta);
             boundaries.second = min*exp((idx+1.0)*delta);
          }
          template <typename T>
          static int verify(const std::vector<T>& mins, const std::vector<T>& maxes,
                        const BoolArray& isLog = BoolArray())
          {
             for (size_t i=0; i<mins.size(); ++i)
             {
                if (mins[i] <= .0 || maxes[i] <=.0 || maxes[i] == mins[i])  
                        return -1;
             }
             return 0;
          }
    };
    
    class MixedBinning
    {
       public:
          template <typename T>
          static int calcIndex(const T& min, const T& max, const T& val, const int nSteps,
                                const bool parIsLog = false)
          {
             if (parIsLog)  return LogBinning::calcIndex(min, max, val, nSteps);
             else  return LinearBinning::calcIndex(min, max, val, nSteps);
          }
          template <typename T>
          static void calcBoundaries(const T& min, const T& max, const int idx, 
                const int nSteps, std::pair<Real,Real>& boundaries, const bool parIsLog = false)
          {
             if (parIsLog) LogBinning::calcBoundaries(min, max, idx, 
                nSteps, boundaries);
             else LinearBinning::calcBoundaries(min, max, idx, 
                nSteps, boundaries);
          }
          template <typename T>
          static int verify(const std::vector<T>& mins, const std::vector<T>& maxes,
                        const BoolArray& isLog = BoolArray())
          {
             if (isLog.size())
             {
                for (size_t i=0; i<mins.size(); ++i)
                {
                   if (isLog[i])
                   {
                      if (mins[i] <= .0 || maxes[i] <=.0 || maxes[i] == mins[i])  
                           return -1;                
                   }
                   else
                   {
                      if (maxes[i] == mins[i]) return -2;
                   }
                }
             }
             else
             {
                for (size_t i=0; i<mins.size(); ++i)
                   if (maxes[i] == mins[i]) return -2;
             }
             return 0;
          }
    };
    //## end Numerics::Histogram%41E6D2050159.preface

    //## Class: Histogram%41E6D2050159; Parameterized Class
    //## Category: XSUtil::Numerics%3F6F5A7603D4
    //## Subsystem: XSUtil::Numerics%41E6D0F3003C
    //## Persistence: Transient
    //## Cardinality/Multiplicity: n



    template <typename T, typename Policy = LinearBinning>
    class Histogram 
    {
      //## begin Numerics::Histogram%41E6D2050159.initialDeclarations preserve=yes
      //## end Numerics::Histogram%41E6D2050159.initialDeclarations

      public:
        //## Constructors (specified)
          //## Operation: Histogram%41E6D644027C
          Histogram (const std::vector<T>& minRanges, const std::vector<T>& maxRanges, const IntegerArray& nSteps, const BoolArray& isLog = BoolArray());

        //## Destructor (generated)
          ~Histogram();


        //## Other Operations (specified)
          //## Operation: nDim%41E6DB780130
          size_t nDim () const;

          //## Operation: addToHist%41E6DCBF0395
          bool addToHist (const std::vector<T>& entry);

          //## Operation: calcBoundaries%41F55ABC02B1
          void calcBoundaries (int index, std::vector<Real>& boundaries) const;

        //## Get and Set Operations for Class Attributes (generated)

          //## Attribute: minRanges%41E6D960025B
          const std::vector<T>& minRanges () const;

          //## Attribute: maxRanges%41E6D99E01BD
          const std::vector<T>& maxRanges () const;

          //## Attribute: nSteps%41E6D9C502F3
          const IntegerArray& nSteps () const;

          //## Attribute: totalEntries%41E6DBD6037A
          long totalEntries () const;

          //## Attribute: counts%41E6DC0D011B
          const std::valarray<long>& counts () const;

          //## Attribute: isLog%41E7E86503A6
          const BoolArray& isLog () const;

          //## Attribute: totalAttempts%41F69A30031F
          long totalAttempts () const;

        // Additional Public Declarations
          //## begin Numerics::Histogram%41E6D2050159.public preserve=yes
          //## end Numerics::Histogram%41E6D2050159.public

      protected:
        // Additional Protected Declarations
          //## begin Numerics::Histogram%41E6D2050159.protected preserve=yes
          //## end Numerics::Histogram%41E6D2050159.protected

      private:
        //## Constructors (generated)
          Histogram(const Histogram< T,Policy > &right);

        //## Assignment Operation (generated)
          Histogram< T,Policy > & operator=(const Histogram< T,Policy > &right);

        // Additional Private Declarations
          //## begin Numerics::Histogram%41E6D2050159.private preserve=yes
          //## end Numerics::Histogram%41E6D2050159.private

      private: //## implementation
        // Data Members for Class Attributes

          //## Attribute: nDim%41E6D86C004E
          //## begin Numerics::Histogram::nDim%41E6D86C004E.attr preserve=no  public: const size_t {U} 
          const size_t m_nDim;
          //## end Numerics::Histogram::nDim%41E6D86C004E.attr

          //## begin Numerics::Histogram::minRanges%41E6D960025B.attr preserve=no  public: std::vector<T> {U} 
          std::vector<T> m_minRanges;
          //## end Numerics::Histogram::minRanges%41E6D960025B.attr

          //## begin Numerics::Histogram::maxRanges%41E6D99E01BD.attr preserve=no  public: std::vector<T> {U} 
          std::vector<T> m_maxRanges;
          //## end Numerics::Histogram::maxRanges%41E6D99E01BD.attr

          //## begin Numerics::Histogram::nSteps%41E6D9C502F3.attr preserve=no  public: IntegerArray {U} 
          IntegerArray m_nSteps;
          //## end Numerics::Histogram::nSteps%41E6D9C502F3.attr

          //## begin Numerics::Histogram::totalEntries%41E6DBD6037A.attr preserve=no  public: long {U} 
          long m_totalEntries;
          //## end Numerics::Histogram::totalEntries%41E6DBD6037A.attr

          //## begin Numerics::Histogram::counts%41E6DC0D011B.attr preserve=no  public: std::valarray<long> {U} 
          std::valarray<long> m_counts;
          //## end Numerics::Histogram::counts%41E6DC0D011B.attr

          //## begin Numerics::Histogram::isLog%41E7E86503A6.attr preserve=no  public: BoolArray {U} 
          BoolArray m_isLog;
          //## end Numerics::Histogram::isLog%41E7E86503A6.attr

          //## begin Numerics::Histogram::totalAttempts%41F69A30031F.attr preserve=no  public: long {U} 
          long m_totalAttempts;
          //## end Numerics::Histogram::totalAttempts%41F69A30031F.attr

        // Additional Implementation Declarations
          //## begin Numerics::Histogram%41E6D2050159.implementation preserve=yes
          //## end Numerics::Histogram%41E6D2050159.implementation

    };

    //## begin Numerics::Histogram%41E6D2050159.postscript preserve=yes
    //## end Numerics::Histogram%41E6D2050159.postscript

    // Parameterized Class Numerics::Histogram 


    //## Other Operations (inline)
    template <typename T, typename Policy>
    inline size_t Histogram<T,Policy>::nDim () const
    {
      //## begin Numerics::Histogram::nDim%41E6DB780130.body preserve=yes
      return m_nDim;
      //## end Numerics::Histogram::nDim%41E6DB780130.body
    }

    //## Get and Set Operations for Class Attributes (inline)

    template <typename T, typename Policy>
    inline const std::vector<T>& Histogram<T,Policy>::minRanges () const
    {
      //## begin Numerics::Histogram::minRanges%41E6D960025B.get preserve=no
      return m_minRanges;
      //## end Numerics::Histogram::minRanges%41E6D960025B.get
    }

    template <typename T, typename Policy>
    inline const std::vector<T>& Histogram<T,Policy>::maxRanges () const
    {
      //## begin Numerics::Histogram::maxRanges%41E6D99E01BD.get preserve=no
      return m_maxRanges;
      //## end Numerics::Histogram::maxRanges%41E6D99E01BD.get
    }

    template <typename T, typename Policy>
    inline const IntegerArray& Histogram<T,Policy>::nSteps () const
    {
      //## begin Numerics::Histogram::nSteps%41E6D9C502F3.get preserve=no
      return m_nSteps;
      //## end Numerics::Histogram::nSteps%41E6D9C502F3.get
    }

    template <typename T, typename Policy>
    inline long Histogram<T,Policy>::totalEntries () const
    {
      //## begin Numerics::Histogram::totalEntries%41E6DBD6037A.get preserve=no
      return m_totalEntries;
      //## end Numerics::Histogram::totalEntries%41E6DBD6037A.get
    }

    template <typename T, typename Policy>
    inline const std::valarray<long>& Histogram<T,Policy>::counts () const
    {
      //## begin Numerics::Histogram::counts%41E6DC0D011B.get preserve=no
      return m_counts;
      //## end Numerics::Histogram::counts%41E6DC0D011B.get
    }

    template <typename T, typename Policy>
    inline const BoolArray& Histogram<T,Policy>::isLog () const
    {
      //## begin Numerics::Histogram::isLog%41E7E86503A6.get preserve=no
      return m_isLog;
      //## end Numerics::Histogram::isLog%41E7E86503A6.get
    }

    template <typename T, typename Policy>
    inline long Histogram<T,Policy>::totalAttempts () const
    {
      //## begin Numerics::Histogram::totalAttempts%41F69A30031F.get preserve=no
      return m_totalAttempts;
      //## end Numerics::Histogram::totalAttempts%41F69A30031F.get
    }

    // Parameterized Class Numerics::Histogram 

    template <typename T, typename Policy>
    Histogram<T,Policy>::Histogram (const std::vector<T>& minRanges, const std::vector<T>& maxRanges, const IntegerArray& nSteps, const BoolArray& isLog)
      //## begin Numerics::Histogram::Histogram%41E6D644027C.hasinit preserve=no
      //## end Numerics::Histogram::Histogram%41E6D644027C.hasinit
      //## begin Numerics::Histogram::Histogram%41E6D644027C.initialization preserve=yes
      : m_nDim(minRanges.size()),
        m_minRanges(minRanges),
        m_maxRanges(maxRanges),
        m_nSteps(nSteps),
        m_totalEntries(0),
        m_counts(),
        m_isLog(),
        m_totalAttempts(0)
      //## end Numerics::Histogram::Histogram%41E6D644027C.initialization
    {
      //## begin Numerics::Histogram::Histogram%41E6D644027C.body preserve=yes
       const string msg("Histogram construction error.");
       if (!m_nDim || m_maxRanges.size() != m_nDim || m_nSteps.size() != m_nDim)
       {
           throw RedAlert(msg);
       }
       
       if (isLog.size())
       {
          if (isLog.size() != m_nDim)   throw RedAlert(msg);
          m_isLog = isLog;
       }
       else
       {
          m_isLog.resize(m_nDim, false);
       }
       size_t nGridPoints = 1;
       for (size_t i=0; i<m_nDim; ++i)
       {
          nGridPoints *= m_nSteps[i];
       }
       if (!nGridPoints)
       {
          throw RedAlert(msg);
       }
       m_counts.resize(nGridPoints, 0);
       int status = Policy::verify(m_minRanges, m_maxRanges, m_isLog);
       if (status == -1)
       {
          const string binErr("Improper limits for log binning.\n");
          throw YellowAlert(binErr);
       }
       else if (status == -2)
       {
          const string binErr("Improper limits for binning.\n");
          throw YellowAlert(binErr);
       }
      //## end Numerics::Histogram::Histogram%41E6D644027C.body
    }


    template <typename T, typename Policy>
    Histogram<T,Policy>::~Histogram()
    {
      //## begin Numerics::Histogram::~Histogram%41E6D2050159_dest.body preserve=yes
      //## end Numerics::Histogram::~Histogram%41E6D2050159_dest.body
    }



    //## Other Operations (implementation)
    template <typename T, typename Policy>
    bool Histogram<T,Policy>::addToHist (const std::vector<T>& entry)
    {
      //## begin Numerics::Histogram::addToHist%41E6DCBF0395.body preserve=yes
       size_t idxTot = 0;
       size_t parIncr = 1;
       ++m_totalAttempts;
       // LOWEST parameter number varies fastest in counts array.
       // Example: if 2 pars, par1 = 2 steps, par2 = 2 steps,
       // counts = [(par1=0,par2=0),(par1=1,par2=0),(par1=0,par2=1),(par1=1,par2=1)] 
       for (size_t i=0; i<m_nDim; ++i)
       {
          bool parIsLog = m_isLog[i];
          int nStep_i = m_nSteps[i];
          int idxPar = Policy::calcIndex(m_minRanges[i], m_maxRanges[i], entry[i], nStep_i,
                                parIsLog);
          if (idxPar < 0 || idxPar >= nStep_i)
          {
             // If ANY indices are out of bounds, point will not get binned.
             return false;
          }
          else
          {
             idxTot += idxPar*parIncr;
             parIncr *= nStep_i;
          }
       }
       m_counts[idxTot] += 1;
       ++m_totalEntries;
       return true;
      //## end Numerics::Histogram::addToHist%41E6DCBF0395.body
    }

    template <typename T, typename Policy>
    void Histogram<T,Policy>::calcBoundaries (int index, std::vector<Real>& boundaries) const
    {
      //## begin Numerics::Histogram::calcBoundaries%41F55ABC02B1.body preserve=yes
       boundaries.resize(m_nDim*2, .0);
       int accum = 1;
       for (size_t i=0; i<m_nDim; ++i)
       {
          int nSteps_i = m_nSteps[i];
          // truncation is deliberate
          int parIdx = (index % (m_nSteps[i]*accum))/accum;
          if (parIdx < 0 || parIdx >= nSteps_i)
          {
             throw YellowAlert("Out of bounds error in Histogram::calcBoundaries\n");
          }
          std::pair<Real, Real> parBnd;
          Policy::calcBoundaries(m_minRanges[i], m_maxRanges[i], parIdx, 
                nSteps_i, parBnd, m_isLog[i]);
          boundaries[2*i] = parBnd.first;
          boundaries[2*i+1] = parBnd.second;
          accum *= m_nSteps[i];
       }
      //## end Numerics::Histogram::calcBoundaries%41F55ABC02B1.body
    }

    // Additional Declarations
      //## begin Numerics::Histogram%41E6D2050159.declarations preserve=yes
      //## end Numerics::Histogram%41E6D2050159.declarations

} // namespace Numerics

//## begin module%41E6D10701BE.epilog preserve=yes
//## end module%41E6D10701BE.epilog


#endif
